import json
import logging
import os
import shutil
import subprocess
import sys
import venv
from types import SimpleNamespace

logger = logging.getLogger(__name__)


def create(  # pylint: disable=R0913
    env_dir,
    system_site_packages=False,
    clear=False,
    symlinks=False,
    upgrade=False,
    with_pip=False,
    prompt=None,
    upgrade_deps=False,
    config_name="default",
):
    """Create a virtual environment in a directory."""
    if sys.version_info[:2] < (3, 9):
        builder = KantanEnvBuilder(
            system_site_packages=system_site_packages,
            clear=clear,
            symlinks=symlinks,
            upgrade=upgrade,
            with_pip=with_pip,
            prompt=prompt,
            config_name=config_name,
        )
    else:
        builder = KantanEnvBuilder(
            system_site_packages=system_site_packages,
            clear=clear,
            symlinks=symlinks,
            upgrade=upgrade,
            with_pip=with_pip,
            prompt=prompt,
            upgrade_deps=upgrade_deps,
            config_name=config_name,
        )

    builder.create(env_dir)


class KantanEnvBuilder(venv.EnvBuilder):

    KANTAN_CONFIG_DIR = os.path.join("~", ".kantan")
    KANTAN_CONFIG_FILE = "configuration.json"

    def __init__(self, *args, **kwargs):
        config_name = kwargs.pop("config_name", "default")
        kantan_home = os.path.expanduser(self.KANTAN_CONFIG_DIR)
        self.config_dir_path = os.path.join(kantan_home, config_name)

        super().__init__(*args, **kwargs)

    def post_setup(self, context: SimpleNamespace) -> None:
        """
        Set up any virtual environment being created based on target
        configuration in KANTAN_CONFIG_DIR.

        :param context: The information for the virtual environment
                        creation request being processed.
        """
        config_file_path = os.path.join(self.config_dir_path, self.KANTAN_CONFIG_FILE)
        if os.path.exists(self.config_dir_path):
            if os.path.isfile(config_file_path):
                with open(config_file_path, "r", encoding="utf8") as config_file:
                    config = json.load(config_file)

                    if "requirements" in config:
                        for package_ref in config["requirements"]:
                            KantanEnvBuilder._install_dependency(context, package_ref)

                    if "include_files" in config:
                        for file_path in config["include_files"]:
                            KantanEnvBuilder._copy_file_to_env_dir(
                                context, self.config_dir_path, file_path
                            )
        else:
            logger.error("No configuration file %s does not exist", config_file_path)

    @staticmethod
    def _install_dependency(context: SimpleNamespace, package_ref: str) -> None:
        logger.debug("Installing %s package in %s", package_ref, context.bin_path)

        cmd = [context.env_exe, "-m", "pip", "install"]
        cmd.append(package_ref)
        subprocess.check_call(cmd)

    @staticmethod
    def _copy_file_to_env_dir(
        context: SimpleNamespace, config_dir_path: str, file_path: str
    ) -> None:
        if not os.path.isabs(file_path):
            file_path = os.path.join(config_dir_path, file_path)
        file_dir, file_name = os.path.split(file_path)
        dst_path = os.path.join(context.env_dir, file_name)

        logger.debug("Copying %s from %s to %s", file_name, file_dir, context.env_dir)
        shutil.copyfile(file_path, dst_path)



import argparse
import os
import sys
import venv

from kantan import builder


def run(args=None):
    if sys.version_info < (3, 7) or not hasattr(sys, "base_prefix"):
        raise ValueError("This script is only for use with Python >= 3.7")

    if sys.version_info[:2] < (3, 9):
        if not hasattr(venv, "CORE_VENV_DEPS"):
            venv.CORE_VENV_DEPS = ('pip', 'setuptools') # type: ignore

    parser = argparse.ArgumentParser(
        prog="kantan",
        description="Creates virtual Python "
        "environments in one or "
        "more target "
        "directories.",
        epilog="Once an environment has been "
        "created, you may wish to "
        "activate it, e.g. by "
        "sourcing an activate script "
        "in its bin directory.",
    )
    parser.add_argument(
        "dirs",
        metavar="ENV_DIR",
        nargs="+",
        help="A directory to create the environment in.",
    )
    parser.add_argument(
        "--system-site-packages",
        default=False,
        action="store_true",
        dest="system_site",
        help="Give the virtual environment access to the " "system site-packages dir.",
    )
    if os.name == "nt":
        use_symlinks = False
    else:
        use_symlinks = True
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        "--symlinks",
        default=use_symlinks,
        action="store_true",
        dest="symlinks",
        help="Try to use symlinks rather than copies, "
        "when symlinks are not the default for "
        "the platform.",
    )
    group.add_argument(
        "--copies",
        default=not use_symlinks,
        action="store_false",
        dest="symlinks",
        help="Try to use copies rather than symlinks, "
        "even when symlinks are the default for "
        "the platform.",
    )
    parser.add_argument(
        "--clear",
        default=False,
        action="store_true",
        dest="clear",
        help="Delete the contents of the "
        "environment directory if it "
        "already exists, before "
        "environment creation.",
    )
    parser.add_argument(
        "--upgrade",
        default=False,
        action="store_true",
        dest="upgrade",
        help="Upgrade the environment "
        "directory to use this version "
        "of Python, assuming Python "
        "has been upgraded in-place.",
    )
    parser.add_argument(
        "--without-pip",
        dest="with_pip",
        default=True,
        action="store_false",
        help="Skips installing or upgrading pip in the "
        "virtual environment (pip is bootstrapped "
        "by default).",
    )
    parser.add_argument(
        "--prompt",
        help="Provides an alternative prompt prefix for " "this environment.",
    )
    parser.add_argument(
        "--upgrade-deps",
        default=False,
        action="store_true",
        dest="upgrade_deps",
        help=f"Upgrade core dependencies: {venv.CORE_VENV_DEPS} to the latest version in PyPI.", # type: ignore
    )
    parser.add_argument(
        "--config-name",
        default="default",
        help="Target configuration to load from `$HOME/.kantan'.",
    )
    options = parser.parse_args(args)
    if options.upgrade and options.clear:
        raise ValueError("you cannot supply --upgrade and --clear together.")

    for directory in options.dirs:
        builder.create(
            directory,
            system_site_packages=options.system_site,
            clear=options.clear,
            symlinks=options.symlinks,
            upgrade=options.upgrade,
            with_pip=options.with_pip,
            prompt=options.prompt,
            upgrade_deps=options.upgrade_deps,
            config_name=options.config_name,
        )


def main():
    return_code = 1
    try:
        run()
        return_code = 0
    except Exception as exception:  # pylint: disable=W0703
        print(f"Error: {exception}", file=sys.stderr)
    sys.exit(return_code)


if __name__ == "__main__":
    main()



import os
import shutil
import sys
from pathlib import Path
from unittest.mock import PropertyMock, patch

from kantan.builder import create
from kantan.cli import run

_tests_root = Path(__file__).parent


def _assert_venv(env_dir: str):
    version_info = sys.version_info
    python_version = f"{version_info[0]}.{version_info[1]}"
    assert os.path.exists(os.path.abspath(env_dir))
    assert os.path.exists(os.path.join(os.path.abspath(env_dir), "pip.conf"))
    assert os.path.exists(
        os.path.join(
            os.path.abspath(env_dir),
            "lib",
            f"python{python_version}",
            "site-packages",
            "kantan",
        )
    )


def _cleanup(env_dir: str):
    shutil.rmtree(os.path.abspath(env_dir))


@patch(
    "kantan.builder.KantanEnvBuilder.KANTAN_CONFIG_DIR",
    new_callable=PropertyMock,
    return_value=f"{_tests_root}",
)
def test_venv_creation(_):
    env_dir = ".test_1"
    create(
        env_dir,
        system_site_packages=False,
        clear=False,
        symlinks=True,
        upgrade=False,
        with_pip=True,
        prompt=None,
        upgrade_deps=False,
        config_name="fixtures",
    )
    _assert_venv(env_dir)
    _cleanup(env_dir)


@patch(
    "kantan.builder.KantanEnvBuilder.KANTAN_CONFIG_DIR",
    new_callable=PropertyMock,
    return_value=f"{_tests_root}",
)
def test_venv_creation_cli(_):
    env_dir = ".test_2"
    run([env_dir, "--config-name", "fixtures"])
    _assert_venv(env_dir)
    _cleanup(env_dir)
